#include "CyanBus.hpp"

// C++ system level
// #include <cstdlib>  // TODO: is this for memcpy?
#include <cinttypes>
#include <functional>

// ESP32 specific
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/uart.h"

// project specific
#include "LambdaTask.hpp"

// qthing stuff
// #include ""



static const char* TAG = "CyanBus";


// This leads to ~54 P/s @ 2 MBaud at 100% load
// TODO: Make configurable?!
const uint32_t CyanBusMTU = 4096;


namespace SiliconTorch {

  namespace CyanBus {


    CyanBus::CyanBus(SiliconTorch::Impl::RS485 reader) : reader(reader) {

      readerTaskHandle = new SiliconTorch::Util::LambdaTask([&](){ readerTask(); });

    }

    void CyanBus::readerTask() {
      const uint8_t* header = (const uint8_t*)"fxCyan";

      uint8_t* buffer = new uint8_t[CyanBusMTU];
      uint8_t* bufPTR = buffer;

      uint16_t packetLength = 0;

      enum FSM { SearchHeader, ReadLenght, ReadBody, CheckCRC };
      FSM fsm = SearchHeader;

      while (true) {

        switch (fsm) {
          case SearchHeader: {

            reader.read(bufPTR, 1);

            if (bufPTR[0] == header[bufPTR - buffer]) {
              // got the right char
              bufPTR++;

              if (bufPTR - buffer >= sizeof(header)) {
                fsm = ReadLenght;
              }
            } else {
              // start over
              bufPTR = buffer;
            }

            break;
          }

          case ReadLenght: {

            reader.read(bufPTR, 2);

            packetLength = (bufPTR[0] << 8) | bufPTR[1];
            bufPTR += 2;

            ESP_LOGW(TAG, "Got a packet with length[ %d ] \\o/", packetLength);

            fsm = ReadBody;

            break;
          }

          case ReadBody: {

            reader.read(bufPTR, packetLength + 4);  // +4 = CRC

            fsm = CheckCRC;

            break;
          }

          case CheckCRC: {
            // TODO: check the CRC!

            // If it matches: handle packet!
            // But if it not matches?
            // search whole packet for header?  (how would we implement this???)
            // or just discard?  (leads to possible desynchronisation!!!)

            break;
          }
        }
      }

      delete buffer;
    }

    void CyanBus::setBaudRate(uint32_t baudRate) {
      ESP_LOGE(TAG, "Operation unsupported: setBaudRate(…)");
    }

    uint32_t CyanBus::getBaudRate() {
      ESP_LOGE(TAG, "Operation unsupported: getBaudRate()");
      return 1337;
    }
  }
}



